# dc3: an experimental, lightweight desktopcouch alternative
# Copyright (C) 2011 Novacut Inc
# 
# This file is part of `dc3`.
# 
# `dc3` is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
# 
# `dc3` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
# 
# You should have received a copy of the GNU Lesser General Public License along
# with `dc3`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
`dc3` - an experimental, lightweight desktopcouch alternative.
"""

import os
from os import path
from subprocess import Popen
import time
import socket
from hashlib import sha1
from base64 import b32encode
import json
import logging

from xdg import BaseDirectory
import microfiber

__version__ = '11.11.0'
NAME = 'dc3'
APPS = '/usr/share/couchdb/apps'
BUS = 'org.freedesktop.DC3'
SOCKET_OPTIONS = '[{recbuf, 262144}, {sndbuf, 262144}, {nodelay, true}]'

log = logging.getLogger()


template = """
[httpd]
bind_address = 127.0.0.1
port = {port}
socket_options = {socket_options}

[couchdb]
view_index_dir = {share}
database_dir = {share}

[log]
file = {couchlog}
level = info

[httpd_global_handlers]
_apps = {apps}
_stats =

[daemons]
stats_collector =
stats_aggregator =

[stats]
rate =
samples =
"""

template_basic = """
[couch_httpd_auth]
require_valid_user = true

[admins]
{username} = {hashed}
"""

template_oauth = """
[oauth_token_users]
{token} = {username}

[oauth_token_secrets]
{token} = {token_secret}

[oauth_consumer_secrets]
{consumer_key} = {consumer_secret}
"""


def handler(d):
    assert path.abspath(d) == d
    return 	'{{couch_httpd_misc_handlers, handle_utils_dir_req, {}}}'.format(
        json.dumps(d)
    )


def random_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('127.0.0.1', 0))
    port = s.getsockname()[1]
    return (s, port)


def logfile(cache, name):
    filename = path.join(cache, name + '.log')
    if path.exists(filename):
        os.rename(filename, filename + '.previous')
    return filename


def mkdir(parent, name):
    d = path.join(parent, name)
    try:
        os.mkdir(d)
    except OSError:
        pass
    return d


def init_dirs(tmp=None):
    if tmp:
        config = mkdir(tmp, 'config')
        share = mkdir(tmp, 'share')
        cache = mkdir(tmp, 'cache')
    else:
        config = BaseDirectory.save_config_path(NAME)
        share = BaseDirectory.save_data_path(NAME)
        cache = mkdir(BaseDirectory.xdg_cache_home, NAME)

    format = [
        '%(levelname)s',
        '%(message)s',
    ]
    logging.basicConfig(
        filename=logfile(cache, 'service'),
        filemode='w',
        level=logging.DEBUG,
        format='\t'.join(format),
    )
    log.info('dc3.__version__: %r', __version__)
    log.info('dc3.__file__: %r', __file__)
    
    ini = path.join(config, 'session.ini')
    d = {
        'share': share,
        'couchlog': logfile(cache, 'couchdb'),
        'apps': handler(APPS),
        'socket_options': SOCKET_OPTIONS,
    }
    return (d, ini)


def random_key():
    return b32encode(os.urandom(10))


def random_oauth():
    return dict(
        (k, random_key())
        for k in ('consumer_key', 'consumer_secret', 'token', 'token_secret')
    )


def random_basic():
    return dict(
        (k, random_key())
        for k in ('username', 'password')
    )

    
def random_env(port):
    return {
        'port': port,
        'url': 'http://localhost:{}/'.format(port),
        #'oauth': random_oauth(),
        #'basic': random_basic(),
    }


def random_salt():
    return os.urandom(16).encode('hex')


def couch_hashed(password, salt):
    hexdigest = sha1(password + salt).hexdigest()
    return '-hashed-{},{}'.format(hexdigest, salt)


def prepare_vars(kw, port):
    env = random_env(port)
    kw['port'] = port
    #kw.update(env['oauth'])
    #kw['username'] = env['basic']['username']
    #kw['hashed'] = couch_hashed(env['basic']['password'], random_salt())
    return env


def prepare_session(port, tmp=None):
    (kw, ini) = init_dirs(tmp)
    env = prepare_vars(kw, port)
    open(ini, 'w').write(template.format(**kw))
    return (env, ini)


def get_cmd(ini):
    return [
        '/usr/bin/couchdb',
        '-n',  # reset configuration file chain (including system default)
        '-a', '/etc/couchdb/default.ini',
        '-a', ini,
    ]


class Monitor(object):
    def __init__(self, env, ini):
        self.server = microfiber.Server(env)
        self.cmd = get_cmd(ini)
        self.couchdb = None

    def start(self):
        assert self.couchdb is None
        log.info('Attempting to start %r', self.server)
        log.info('Command: %r', self.cmd)
        self.couchdb = Popen(self.cmd)
        time.sleep(0.35)
        t = 0.2
        for i in range(10):
            if self.isalive():
                return True
            time.sleep(t)
            t += 0.1
        return True
        raise Exception('could not start CouchDB')

    def isalive(self):
        try:
            self.server.get()
            log.info('%r is alive', self.server)
            return True
        except socket.error:
            log.warning('%r is NOT alive', self.server)
            return False

    def kill(self):
        if self.couchdb is None:
            return
        log.info('Terminating %r', self.server)
        self.couchdb.terminate()
        self.couchdb.wait()
        self.couchdb = None

    def check(self):
        if not self.isalive():
            log.error('%s has failed, restarting...', self.server)
            self.kill()
            self.start()

    def crash(self):
        """
        Do a soft kill for testing purposes to see if it is re-spawned.
        """
        if self.couchdb is None:
            return
        log.warning('Crashing %r', self.server)
        self.couchdb.terminate()

    def set_intree(self, treedir):
        assert path.abspath(treedir) == treedir
        assert path.isdir(treedir)
        return self.server.put(
            handler(treedir), '_config', 'httpd_global_handlers', '_intree'
        )

    def del_intree(self):
        return self.server.delete('_config', 'httpd_global_handlers', '_intree')
